(* "by rec" *)

Require Import Equations.Equations.
Require Import Arith Omega.
Require Import Relations.
Require Import List Sorting.
Import Permutation.
Import ListNotations.

Section QuickSort.
  Variable A : Type.
  Variable lt_A : A -> A -> Prop.
  Variable lt_A_dec : forall (x y : A),
    {lt_A x y} + {~ lt_A x y}.

  Equations partition (p : A) (l : list A) :
    list A * list A :=
  partition p nil := (nil, nil);
  partition p (cons hd tl) <= partition p tl => {
    | pair l1 l2 <= lt_A_dec hd p => {
      | left H := (hd :: l1, l2);
      | right H := (l1, hd :: l2)
    }
  }.

  (* Some post-condition. *)
  Definition is_partition (p : A) (l : list A)
    (res : list A * list A) : Prop :=
    let (l1, l2) := res in
    Forall (fun x => lt_A x p) l1 /\
    Forall (fun x => ~ lt_A x p) l2 /\
    Permutation l (l1 ++ l2).

  Lemma partition_post (p : A) (l : list A) :
    is_partition p l (partition p l).
  Proof.
    funelim (partition p l); simpl.
    - intuition.
    - rewrite Heq0 in Hind;
      simpl in Hind; intuition.
    - rewrite Heq0 in Hind; simpl in Hind; 
      intuition eauto using Permutation_cons_app.
  Qed.

  Definition partition' (p : A) (l : list A) :
    {res : list A * list A | is_partition p l res} :=
    exist _ (partition p l) (partition_post p l).

  Lemma partition_length (p : A) (l : list A) :
    forall l1 l2,
    is_partition p l (l1, l2) ->
    length l = length l1 + length l2.
  Proof. intros. destruct H as [_ [_ H]].
   rewrite (Permutation_length H). apply app_length.
  Qed.

  Definition order : relation (list A) :=
    MR lt (@length A).

  Theorem wf_order : well_founded order.
  Proof.
    apply wf_MR. apply lt_wf.
  Qed.

  Lemma partition_order_1 (p : A) (l : list A) :
    forall l1 l2,
    is_partition p l (l1, l2) ->
    order l1 (p :: l).
  Proof. intros.
    do 2 red. simpl.
    rewrite (partition_length _ _ _ _ H); omega.
  Qed.

  Lemma partition_order_2 (p : A) (l : list A) :
    forall l1 l2,
    is_partition p l (l1, l2) ->
    order l2 (p :: l).
  Proof. intros.
    do 2 red. simpl.
    rewrite (partition_length _ _ _ _ H); omega.
  Qed.
  Hint Resolve partition_order_1.
  Hint Resolve partition_order_2.

  Equations quicksort (l : list A) : list A :=
  quicksort l by rec l order :=
  quicksort nil := nil;
  quicksort (cons p tl) <= partition' p tl => {
    | exist (pair l1 l2) H :=
      quicksort l1 ++ p :: quicksort l2
  }.

  Obligation Tactic := idtac.
  Next Obligation.
    intros; apply quicksort; eauto.
  Defined.
  Next Obligation.
    intros; apply quicksort; eauto.
  Defined.

  Next Obligation.
    apply (@FixWf _ _ wf_order).
    intros.
    depelim x.
    - constructor.
    - simp quicksort. constructor.
      destruct (partition' a x). simpl.
      depelim x0. constructor; eauto.
  Defined.
End QuickSort.

Compute quicksort nat lt lt_dec
  [4; 4; 2; 1; 10; 3; 7].

(* Extraction. *)
Extraction Inline partition_obligation_1
  partition_obligation_2 partition_obligation_3
  partition_obligation_4 partition_obligation_5
  quicksort_obligation_1 quicksort_obligation_2
  quicksort_obligation_3 quicksort_obligation_4
  quicksort_obligation_5 quicksort_obligation_6
  quicksort_comp_proj.
Extract Inductive list => "list" [ "[]" "(::)" ].
Extract Inductive prod => "(*)" [ "(,)" ].
Extract Inductive sumbool => "bool" ["true" "false"].
Extract Inductive unit => "()" ["()"].
Extract Inductive sigma => "(*)" [ "(,)" ].
Extract Inlined Constant pr1 => "fst".
Extract Inlined Constant app => "(@)".
Recursive Extraction quicksort.